from typing import Any
from typing import Dict
from typing import List
from typing import Union
from typing import Optional


general_config_type = Optional[Union[str, Dict[str, Any]]]
configs_type = Optional[Union[List[Dict[str, Any]], Dict[str, Any]]]
